@@ -5,17 +5,20 @@ from functools import cached_property
 from typing_extensions import Self
 
 from pydantic import Field
-from sqlglot import Dialect, exp
+from sqlglot import Dialect, exp, parse_one
 from sqlglot.helper import ensure_collection, ensure_list
 from sqlglot.optimizer.normalize_identifiers import normalize_identifiers
 
 from sqlmesh.core import dialect as d
-from sqlmesh.core.dialect import normalize_model_name, extract_func_call
+from sqlmesh.core.config.common import VirtualEnvironmentMode
+from sqlmesh.core.config.linter import LinterConfig
+from sqlmesh.core.dialect import normalize_model_name
 from sqlmesh.core.model.common import (
     bool_validator,
     default_catalog_validator,
     depends_on_validator,
     properties_validator,
+    parse_properties,
 )
 from sqlmesh.core.model.kind import (
     CustomKind,
@@ -26,8 +29,8 @@ from sqlmesh.core.model.kind import (
     SCDType2ByTimeKind,
     TimeColumn,
     ViewKind,
-    _IncrementalBy,
     model_kind_validator,
+    OnAdditiveChange,
 )
 from sqlmesh.core.node import _Node, str_or_exp_to_str
 from sqlmesh.core.reference import Reference
@@ -38,6 +41,7 @@ from sqlmesh.utils.pydantic import (
     field_validator,
     list_of_fields_validator,
     model_validator,
+    get_dialect,
 )
 
 if t.TYPE_CHECKING:
@@ -76,7 +80,11 @@ class ModelMeta(_Node):
     physical_version: t.Optional[str] = None
     gateway: t.Optional[str] = None
     optimize_query: t.Optional[bool] = None
-    validate_query: t.Optional[bool] = None
+    ignored_rules_: t.Optional[t.Set[str]] = Field(
+        default=None, exclude=True, alias="ignored_rules"
+    )
+    formatting: t.Optional[bool] = Field(default=None, exclude=True)
+    virtual_environment_mode: VirtualEnvironmentMode = VirtualEnvironmentMode.default
 
     _bool_validator = bool_validator
     _model_kind_validator = model_kind_validator
@@ -88,37 +96,7 @@ class ModelMeta(_Node):
     def _func_call_validator(cls, v: t.Any, field: t.Any) -> t.Any:
         is_signal = getattr(field, "name" if hasattr(field, "name") else "field_name") == "signals"
 
-        if isinstance(v, (exp.Tuple, exp.Array)):
-            return [extract_func_call(i, allow_tuples=is_signal) for i in v.expressions]
-        if isinstance(v, exp.Paren):
-            return [extract_func_call(v.this, allow_tuples=is_signal)]
-        if isinstance(v, exp.Expression):
-            return [extract_func_call(v, allow_tuples=is_signal)]
-        if isinstance(v, list):
-            audits = []
-
-            for entry in v:
-                if isinstance(entry, dict):
-                    args = entry
-                    name = "" if is_signal else entry.pop("name")
-                elif isinstance(entry, (tuple, list)):
-                    name, args = entry
-                else:
-                    raise ConfigError(f"Audit must be a dictionary or named tuple. Got {entry}.")
-
-                audits.append(
-                    (
-                        name.lower(),
-                        {
-                            key: d.parse_one(value) if isinstance(value, str) else value
-                            for key, value in args.items()
-                        },
-                    )
-                )
-
-            return audits
-
-        return v or []
+        return d.extract_function_calls(v, allow_tuples=is_signal)
 
     @field_validator("tags", mode="before")
     def _value_or_tuple_validator(cls, v: t.Any, info: ValidationInfo) -> t.Any:
@@ -178,6 +156,22 @@ class ModelMeta(_Node):
     def _partition_and_cluster_validator(
         cls, v: t.Any, info: ValidationInfo
     ) -> t.List[exp.Expression]:
+        if (
+            isinstance(v, list)
+            and all(isinstance(i, str) for i in v)
+            and info.field_name == "partitioned_by_"
+        ):
+            # this branch gets hit when we are deserializing from json because `partitioned_by` is stored as a List[str]
+            # however, we should only invoke this if the list contains strings because this validator is also
+            # called by Python models which might pass a List[exp.Expression]
+            string_to_parse = (
+                f"({','.join(v)})"  # recreate the (a, b, c) part of "partitioned_by (a, b, c)"
+            )
+            parsed = parse_one(
+                string_to_parse, into=exp.PartitionedByProperty, dialect=get_dialect(info)
+            )
+            v = parsed.this.expressions if isinstance(parsed.this, exp.Schema) else v
+
         expressions = list_of_fields_validator(v, info.data)
 
         for expression in expressions:
@@ -205,7 +199,10 @@ class ModelMeta(_Node):
 
         if isinstance(v, exp.Schema):
             for column in v.expressions:
-                expr = column.args["kind"]
+                expr = column.args.get("kind")
+                if not isinstance(expr, exp.DataType):
+                    raise ConfigError(f"Missing data type for column '{column.name}'.")
+
                 expr.meta["dialect"] = dialect
                 columns_to_types[normalize_identifiers(column, dialect=dialect).name] = expr
 
@@ -282,6 +279,55 @@ class ModelMeta(_Node):
 
         return refs
 
+    @field_validator("ignored_rules_", mode="before")
+    def ignored_rules_validator(cls, vs: t.Any) -> t.Any:
+        return LinterConfig._validate_rules(vs)
+
+    @field_validator("session_properties_", mode="before")
+    def session_properties_validator(cls, v: t.Any, info: ValidationInfo) -> t.Any:
+        # use the generic properties validator to parse the session properties
+        parsed_session_properties = parse_properties(type(cls), v, info)
+        if not parsed_session_properties:
+            return parsed_session_properties
+
+        for eq in parsed_session_properties:
+            prop_name = eq.left.name
+
+            if prop_name == "query_label":
+                query_label = eq.right
+                if not isinstance(
+                    query_label, (exp.Array, exp.Tuple, exp.Paren, d.MacroFunc, d.MacroVar)
+                ):
+                    raise ConfigError(
+                        "Invalid value for `session_properties.query_label`. Must be an array or tuple."
+                    )
+
+                label_tuples: t.List[exp.Expression] = (
+                    [query_label.unnest()]
+                    if isinstance(query_label, exp.Paren)
+                    else query_label.expressions
+                )
+
+                for label_tuple in label_tuples:
+                    if not (
+                        isinstance(label_tuple, exp.Tuple)
+                        and len(label_tuple.expressions) == 2
+                        and all(isinstance(label, exp.Literal) for label in label_tuple.expressions)
+                    ):
+                        raise ConfigError(
+                            "Invalid entry in `session_properties.query_label`. Must be tuples of string literals with length 2."
+                        )
+            elif prop_name == "authorization":
+                authorization = eq.right
+                if not (
+                    isinstance(authorization, exp.Literal) and authorization.is_string
+                ) and not isinstance(authorization, (d.MacroFunc, d.MacroVar)):
+                    raise ConfigError(
+                        "Invalid value for `session_properties.authorization`. Must be a string literal."
+                    )
+
+        return parsed_session_properties
+
     @model_validator(mode="before")
     def _pre_root_validator(cls, data: t.Any) -> t.Any:
         if not isinstance(data, dict):
@@ -327,7 +373,7 @@ class ModelMeta(_Node):
                 and not (kind.is_view and kind.materialized)
             ):
                 name = field[:-1] if field.endswith("_") else field
-                raise ValueError(f"{name} field cannot be set for {kind} models")
+                raise ValueError(f"{name} field cannot be set for {kind.name} models")
         if kind.is_incremental_by_partition and not getattr(self, "partitioned_by_", None):
             raise ValueError(f"partitioned_by field is required for {kind.name} models")
 
@@ -367,7 +413,7 @@ class ModelMeta(_Node):
     @property
     def lookback(self) -> int:
         """The incremental lookback window."""
-        return (self.kind.lookback if isinstance(self.kind, _IncrementalBy) else 0) or 0
+        return getattr(self.kind, "lookback", 0) or 0
 
     def lookback_start(self, start: TimeLike) -> TimeLike:
         if self.lookback == 0:
@@ -422,6 +468,20 @@ class ModelMeta(_Node):
             Reference(model_name=self.name, expression=e, unique=True) for e in self.references
         ]
 
+    @property
+    def on(self) -> t.List[str]:
+        """The grains to be used as join condition in table_diff."""
+
+        on: t.List[str] = []
+        for expr in [ref.expression for ref in self.all_references if ref.unique]:
+            if isinstance(expr, exp.Tuple):
+                on.extend([key.this.sql(dialect=self.dialect) for key in expr.expressions])
+            else:
+                # Handle a single Column or Paren expression
+                on.append(expr.this.sql(dialect=self.dialect))
+
+        return on
+
     @property
     def managed_columns(self) -> t.Dict[str, exp.DataType]:
         return getattr(self.kind, "managed_columns", {})
@@ -456,3 +516,12 @@ class ModelMeta(_Node):
     @property
     def on_destructive_change(self) -> OnDestructiveChange:
         return getattr(self.kind, "on_destructive_change", OnDestructiveChange.ALLOW)
+
+    @property
+    def on_additive_change(self) -> OnAdditiveChange:
+        """Return the model's additive change setting if it has one."""
+        return getattr(self.kind, "on_additive_change", OnAdditiveChange.ALLOW)
+
+    @property
+    def ignored_rules(self) -> t.Set[str]:
+        return self.ignored_rules_ or set()
