@@ -1,18 +1,19 @@
 from __future__ import annotations
 
+import sys
+
 import datetime
+import threading
 import typing as t
 import unittest
 from collections import Counter
-from contextlib import AbstractContextManager, nullcontext
+from contextlib import nullcontext, contextmanager, AbstractContextManager
+from itertools import chain
 from pathlib import Path
 from unittest.mock import patch
 
-import numpy as np
-import pandas as pd
+
 from io import StringIO
-from freezegun import freeze_time
-from pandas.api.types import is_object_dtype
 from sqlglot import Dialect, exp
 from sqlglot.optimizer.annotate_types import annotate_types
 from sqlglot.optimizer.normalize_identifiers import normalize_identifiers
@@ -23,19 +24,35 @@ from sqlmesh.core.engine_adapter import EngineAdapter
 from sqlmesh.core.macros import RuntimeStage
 from sqlmesh.core.model import Model, PythonModel, SqlModel
 from sqlmesh.utils import UniqueKeyDict, random_id, type_is_known, yaml
-from sqlmesh.utils.date import pandas_timestamp_to_pydatetime
+from sqlmesh.utils.date import date_dict, pandas_timestamp_to_pydatetime, to_datetime
 from sqlmesh.utils.errors import ConfigError, TestError
 from sqlmesh.utils.yaml import load as yaml_load
+from sqlmesh.utils import Verbosity
+from sqlmesh.utils.rich import df_to_table
 
 if t.TYPE_CHECKING:
+    import pandas as pd
+
     from sqlglot.dialects.dialect import DialectType
 
     Row = t.Dict[str, t.Any]
 
 
+TIME_KWARG_KEYS = {
+    "start",
+    "end",
+    "execution_time",
+    "latest",
+    # all built-in datetime macro var names
+    *date_dict(execution_time="1970-01-01", start="1970-01-01", end="1970-01-01").keys(),
+}
+
+
 class ModelTest(unittest.TestCase):
     __test__ = False
 
+    CONCURRENT_RENDER_LOCK = threading.Lock()
+
     def __init__(
         self,
         body: t.Dict[str, t.Any],
@@ -47,6 +64,8 @@ class ModelTest(unittest.TestCase):
         path: Path | None = None,
         preserve_fixtures: bool = False,
         default_catalog: str | None = None,
+        concurrency: bool = False,
+        verbosity: Verbosity = Verbosity.DEFAULT,
     ) -> None:
         """ModelTest encapsulates a unit test for a model.
 
@@ -69,6 +88,8 @@ class ModelTest(unittest.TestCase):
         self.preserve_fixtures = preserve_fixtures
         self.default_catalog = default_catalog
         self.dialect = dialect
+        self.concurrency = concurrency
+        self.verbosity = verbosity
 
         self._fixture_table_cache: t.Dict[str, exp.Table] = {}
         self._normalized_column_name_cache: t.Dict[str, str] = {}
@@ -85,54 +106,77 @@ class ModelTest(unittest.TestCase):
         else:
             self._fixture_catalog = None
 
-        # The test schema name is randomized to avoid concurrency issues
-        self._fixture_schema = exp.to_identifier(f"sqlmesh_test_{random_id(short=True)}")
+        # The test schema name is randomized to avoid concurrency issues,
+        # unless a schema is provided in the unit tests's body
+        self._fixture_schema = exp.parse_identifier(
+            self.body.get("schema") or f"sqlmesh_test_{random_id(short=True)}"
+        )
         self._qualified_fixture_schema = schema_(self._fixture_schema, self._fixture_catalog)
 
         self._transforms = self._test_adapter_dialect.generator_class.TRANSFORMS
         self._execution_time = str(self.body.get("vars", {}).get("execution_time") or "")
 
+        if self._execution_time:
+            # Normalizes the execution time by converting it into UTC timezone
+            self._execution_time = str(to_datetime(self._execution_time))
+
         # When execution_time is set, we mock the CURRENT_* SQL expressions so they always return it
         if self._execution_time:
             exec_time = exp.Literal.string(self._execution_time)
             self._transforms = {
                 **self._transforms,
-                exp.CurrentDate: lambda self, _: self.sql(exp.cast(exec_time, "date")),
-                exp.CurrentDatetime: lambda self, _: self.sql(exp.cast(exec_time, "datetime")),
-                exp.CurrentTime: lambda self, _: self.sql(exp.cast(exec_time, "time")),
-                exp.CurrentTimestamp: lambda self, _: self.sql(exp.cast(exec_time, "timestamp")),
+                exp.CurrentDate: lambda self, _: self.sql(
+                    exp.cast(exec_time, "date", dialect=dialect)
+                ),
+                exp.CurrentDatetime: lambda self, _: self.sql(
+                    exp.cast(exec_time, "datetime", dialect=dialect)
+                ),
+                exp.CurrentTime: lambda self, _: self.sql(
+                    exp.cast(exec_time, "time", dialect=dialect)
+                ),
+                exp.CurrentTimestamp: lambda self, _: self.sql(
+                    exp.cast(exec_time, "timestamp", dialect=dialect)
+                ),
             }
 
         super().__init__()
 
+    def defaultTestResult(self) -> unittest.TestResult:
+        from sqlmesh.core.test.result import ModelTextTestResult
+
+        return ModelTextTestResult(stream=sys.stdout, descriptions=True, verbosity=self.verbosity)
+
     def shortDescription(self) -> t.Optional[str]:
         return self.body.get("description")
 
     def setUp(self) -> None:
         """Load all input tables"""
+        import pandas as pd
+        import numpy as np
+
         self.engine_adapter.create_schema(self._qualified_fixture_schema)
 
         for name, values in self.body.get("inputs", {}).items():
             all_types_are_known = False
-            known_columns_to_types: t.Dict[str, exp.DataType] = {}
+            columns_to_known_types: t.Dict[str, exp.DataType] = {}
 
             model = self.models.get(name)
             if model:
                 inferred_columns_to_types = model.columns_to_types or {}
-                known_columns_to_types = {
+                columns_to_known_types = {
                     c: t for c, t in inferred_columns_to_types.items() if type_is_known(t)
                 }
                 all_types_are_known = bool(inferred_columns_to_types) and (
-                    len(known_columns_to_types) == len(inferred_columns_to_types)
+                    len(columns_to_known_types) == len(inferred_columns_to_types)
                 )
 
             # Types specified in the test will override the corresponding inferred ones
-            known_columns_to_types.update(values.get("columns", {}))
+            columns_to_known_types.update(values.get("columns", {}))
 
             rows = values.get("rows")
             if not all_types_are_known and rows:
                 for col, value in rows[0].items():
-                    if col not in known_columns_to_types:
+                    if col not in columns_to_known_types:
                         v_type = annotate_types(exp.convert(value)).type or type(value).__name__
                         v_type = exp.maybe_parse(
                             v_type, into=exp.DataType, dialect=self._test_adapter_dialect
@@ -147,17 +191,25 @@ class ModelTest(unittest.TestCase):
                                 self.path,
                             )
 
-                        known_columns_to_types[col] = v_type
+                        columns_to_known_types[col] = v_type
 
             if rows is None:
                 query_or_df: exp.Query | pd.DataFrame = self._add_missing_columns(
-                    values["query"], known_columns_to_types
+                    values["query"], columns_to_known_types
                 )
+                if columns_to_known_types:
+                    columns_to_known_types = {
+                        col: columns_to_known_types[col] for col in query_or_df.named_selects
+                    }
             else:
-                query_or_df = self._create_df(values, columns=known_columns_to_types)
+                query_or_df = self._create_df(values, columns=columns_to_known_types)
+
+            # Convert NaN/NaT values to None if DataFrame
+            if isinstance(query_or_df, pd.DataFrame):
+                query_or_df = query_or_df.replace({np.nan: None})
 
             self.engine_adapter.create_view(
-                self._test_fixture_table(name), query_or_df, known_columns_to_types
+                self._test_fixture_table(name), query_or_df, columns_to_known_types
             )
 
     def tearDown(self) -> None:
@@ -173,9 +225,13 @@ class ModelTest(unittest.TestCase):
         partial: t.Optional[bool] = False,
     ) -> None:
         """Compare two DataFrames"""
+        import numpy as np
+        import pandas as pd
+        from pandas.api.types import is_object_dtype
+
         if partial:
             intersection = actual[actual.columns.intersection(expected.columns)]
-            if not intersection.empty:
+            if len(intersection.columns) > 0:
                 actual = intersection
 
         # Two astypes are necessary, pandas converts strings to times as NS,
@@ -202,26 +258,46 @@ class ModelTest(unittest.TestCase):
             if is_object_dtype(actual_types[col]) and len(actual[col]) != 0
         }
         for col, value in object_sentinel_values.items():
-            # can't use `isinstance()` here - https://stackoverflow.com/a/68743663/1707525
-            if type(value) is datetime.date:
-                expected[col] = pd.to_datetime(expected[col], errors="ignore").dt.date  # type: ignore
-            elif type(value) is datetime.time:
-                expected[col] = pd.to_datetime(expected[col], errors="ignore").dt.time  # type: ignore
-            elif type(value) is datetime.datetime:
-                expected[col] = pd.to_datetime(expected[col], errors="ignore").dt.to_pydatetime()  # type: ignore
+            try:
+                # can't use `isinstance()` here - https://stackoverflow.com/a/68743663/1707525
+                if type(value) is datetime.date:
+                    expected[col] = pd.to_datetime(expected[col]).dt.date
+                elif type(value) is datetime.time:
+                    expected[col] = pd.to_datetime(expected[col]).dt.time
+                elif type(value) is datetime.datetime:
+                    expected[col] = pd.to_datetime(expected[col]).dt.to_pydatetime()
+            except Exception as e:
+                from sqlmesh.core.console import get_console
+
+                get_console().log_warning(
+                    f"Failed to convert expected value for {col} into `datetime` "
+                    f"for unit test '{str(self)}'. {str(e)}."
+                )
 
         actual = actual.replace({np.nan: None})
         expected = expected.replace({np.nan: None})
 
+        # We define this here to avoid a top-level import of numpy and pandas
+        DATETIME_TYPES = (
+            datetime.datetime,
+            datetime.date,
+            datetime.time,
+            np.datetime64,
+            pd.Timestamp,
+        )
+
         def _to_hashable(x: t.Any) -> t.Any:
             if isinstance(x, (list, np.ndarray)):
-                return tuple(x)
-            return str(x) if not isinstance(x, t.Hashable) else x
+                return tuple(_to_hashable(v) for v in x)
+            if isinstance(x, dict):
+                return tuple((k, _to_hashable(v)) for k, v in x.items())
+            return str(x) if isinstance(x, DATETIME_TYPES) or not isinstance(x, t.Hashable) else x
+
+        actual = actual.apply(lambda col: col.map(_to_hashable))
+        expected = expected.apply(lambda col: col.map(_to_hashable))
 
         if sort:
-            actual = actual.apply(lambda col: col.map(_to_hashable))
             actual = actual.sort_values(by=actual.columns.to_list()).reset_index(drop=True)
-            expected = expected.apply(lambda col: col.map(_to_hashable))
             expected = expected.sort_values(by=expected.columns.to_list()).reset_index(drop=True)
 
         try:
@@ -233,23 +309,62 @@ class ModelTest(unittest.TestCase):
                 check_like=True,  # Ignore column order
             )
         except AssertionError as e:
+            # There are 2 concepts at play here:
+            # 1. The Exception args will contain the error message plus the diff dataframe table stringified
+            #    (backwards compatibility with existing tests, possible to serialize/send over network etc)
+            # 2. Each test will also transform these diff dataframes into Rich tables, which will be the ones that'll
+            #    be surfaced to the user through Console for better UX (versus stringified dataframes)
+            #
+            # This is a bit of a hack, but it's a way to get the best of both worlds.
+            args: t.List[t.Any] = []
+
+            failed_subtest = ""
+
+            if subtest := getattr(self, "_subtest", None):
+                if cte := subtest.params.get("cte"):
+                    failed_subtest = f" (CTE {cte})"
+
             if expected.shape != actual.shape:
                 _raise_if_unexpected_columns(expected.columns, actual.columns)
 
-                error_msg = "Data mismatch (rows are different)"
+                args.append("Data mismatch (rows are different)")
 
                 missing_rows = _row_difference(expected, actual)
                 if not missing_rows.empty:
-                    error_msg += f"\n\nMissing rows:\n\n{missing_rows}"
+                    args[0] += f"\n\nMissing rows:\n\n{missing_rows}"
+                    args.append(df_to_table(f"Missing rows{failed_subtest}", missing_rows))
 
                 unexpected_rows = _row_difference(actual, expected)
-                if not unexpected_rows.empty:
-                    error_msg += f"\n\nUnexpected rows:\n\n{unexpected_rows}"
 
-                e.args = (error_msg,)
+                if not unexpected_rows.empty:
+                    args[0] += f"\n\nUnexpected rows:\n\n{unexpected_rows}"
+                    args.append(df_to_table(f"Unexpected rows{failed_subtest}", unexpected_rows))
+
             else:
                 diff = expected.compare(actual).rename(columns={"self": "exp", "other": "act"})
-                e.args = (f"Data mismatch (exp: expected, act: actual)\n\n{diff}",)
+
+                args.append(f"Data mismatch (exp: expected, act: actual)\n\n{diff}")
+
+                diff.rename(columns={"exp": "Expected", "act": "Actual"}, inplace=True)
+                if self.verbosity == Verbosity.DEFAULT:
+                    args.extend(
+                        df_to_table(f"Data mismatch{failed_subtest}", df)
+                        for df in _split_df_by_column_pairs(diff)
+                    )
+                else:
+                    from pandas import MultiIndex
+
+                    levels = t.cast(MultiIndex, diff.columns).levels[0]
+                    for col in levels:
+                        col_diff = diff[col]
+                        if not col_diff.empty:
+                            table = df_to_table(
+                                f"[bold red]Column '{col}' mismatch{failed_subtest}[/bold red]",
+                                col_diff,
+                            )
+                            args.append(table)
+
+            e.args = (*args,)
 
             raise e
 
@@ -270,7 +385,9 @@ class ModelTest(unittest.TestCase):
         path: Path | None,
         preserve_fixtures: bool = False,
         default_catalog: str | None = None,
-    ) -> ModelTest:
+        concurrency: bool = False,
+        verbosity: Verbosity = Verbosity.DEFAULT,
+    ) -> t.Optional[ModelTest]:
         """Create a SqlModelTest or a PythonModelTest.
 
         Args:
@@ -282,10 +399,19 @@ class ModelTest(unittest.TestCase):
             path: An optional path to the test definition yaml file.
             preserve_fixtures: Preserve the fixture tables in the testing database, useful for debugging.
         """
-        name = normalize_model_name(body["model"], default_catalog=default_catalog, dialect=dialect)
+        name = body.get("model")
+        if name is None:
+            _raise_error("Missing required 'model' field", path)
+
+        name = normalize_model_name(name, default_catalog=default_catalog, dialect=dialect)
         model = models.get(name)
         if not model:
-            _raise_error(f"Model '{name}' was not found", path)
+            from sqlmesh.core.console import get_console
+
+            get_console().log_warning(
+                f"Model '{name}' was not found{' at ' + str(path) if path else ''}"
+            )
+            return None
 
         if isinstance(model, SqlModel):
             test_type: t.Type[ModelTest] = SqlModelTest
@@ -294,17 +420,22 @@ class ModelTest(unittest.TestCase):
         else:
             _raise_error(f"Model '{name}' is an unsupported model type for testing", path)
 
-        return test_type(
-            body,
-            test_name,
-            t.cast(Model, model),
-            models,
-            engine_adapter,
-            dialect,
-            path,
-            preserve_fixtures,
-            default_catalog,
-        )
+        try:
+            return test_type(
+                body,
+                test_name,
+                t.cast(Model, model),
+                models,
+                engine_adapter,
+                dialect,
+                path,
+                preserve_fixtures,
+                default_catalog,
+                concurrency,
+                verbosity,
+            )
+        except Exception as e:
+            raise TestError(f"Failed to create test {test_name} ({path})\n{str(e)}")
 
     def __str__(self) -> str:
         return f"{self.test_name} ({self.path})"
@@ -326,6 +457,8 @@ class ModelTest(unittest.TestCase):
             partial: bool = False,
             dialect: DialectType = None,
         ) -> t.Dict:
+            import pandas as pd
+
             if not isinstance(values, dict):
                 values = {"rows": values}
 
@@ -463,10 +596,34 @@ class ModelTest(unittest.TestCase):
 
         return normalized_name
 
-    def _execute(self, query: exp.Query) -> pd.DataFrame:
+    @contextmanager
+    def _concurrent_render_context(self) -> t.Iterator[None]:
+        """
+        Context manager that ensures that the tests are executed safely in a concurrent environment.
+        This is needed in case `execution_time` is set, as we'd then have to:
+        - Freeze time through `time_machine` (not thread safe)
+        - Globally patch the SQLGlot dialect so that any date/time nodes are evaluated at the `execution_time` during generation
+        """
+        import time_machine
+
+        lock_ctx: AbstractContextManager = (
+            self.CONCURRENT_RENDER_LOCK if self.concurrency else nullcontext()
+        )
+        time_ctx: AbstractContextManager = nullcontext()
+        dialect_patch_ctx: AbstractContextManager = nullcontext()
+
+        if self._execution_time:
+            time_ctx = time_machine.travel(self._execution_time, tick=False)
+            dialect_patch_ctx = patch.dict(
+                self._test_adapter_dialect.generator_class.TRANSFORMS, self._transforms
+            )
+
+        with lock_ctx, time_ctx, dialect_patch_ctx:
+            yield
+
+    def _execute(self, query: exp.Query | str) -> pd.DataFrame:
         """Executes the given query using the testing engine adapter and returns a DataFrame."""
-        with patch.dict(self._test_adapter_dialect.generator_class.TRANSFORMS, self._transforms):
-            return self.engine_adapter.fetchdf(query)
+        return self.engine_adapter.fetchdf(query)
 
     def _create_df(
         self,
@@ -474,9 +631,14 @@ class ModelTest(unittest.TestCase):
         columns: t.Optional[t.Collection] = None,
         partial: t.Optional[bool] = False,
     ) -> pd.DataFrame:
+        import pandas as pd
+
         query = values.get("query")
         if query:
-            return self._execute(self._add_missing_columns(query, columns))
+            if not partial:
+                query = self._add_missing_columns(query, columns)
+
+            return self._execute(query)
 
         rows = values["rows"]
         if columns:
@@ -486,7 +648,9 @@ class ModelTest(unittest.TestCase):
             if partial:
                 columns = referenced_columns
 
-        return pd.DataFrame.from_records(rows, columns=columns)
+        return pd.DataFrame.from_records(
+            rows, columns=[str(c) for c in columns] if columns else None
+        )
 
     def _add_missing_columns(
         self, query: exp.Query, all_columns: t.Optional[t.Collection[str]] = None
@@ -503,7 +667,7 @@ class ModelTest(unittest.TestCase):
 
 
 class SqlModelTest(ModelTest):
-    def test_ctes(self, ctes: t.Dict[str, exp.Expression]) -> None:
+    def test_ctes(self, ctes: t.Dict[str, exp.Expression], recursive: bool = False) -> None:
         """Run CTE queries and compare output to expected output"""
         for cte_name, values in self.body["outputs"].get("ctes", {}).items():
             with self.subTest(cte=cte_name):
@@ -513,46 +677,68 @@ class SqlModelTest(ModelTest):
                     )
 
                 cte_query = ctes[cte_name].this
-                for alias, cte in ctes.items():
-                    cte_query = cte_query.with_(alias, cte.this)
 
-                partial = values.get("partial")
                 sort = cte_query.args.get("order") is None
+                partial = values.get("partial")
 
-                actual = self._execute(cte_query)
+                cte_query = exp.select(*_projection_identifiers(cte_query)).from_(cte_name)
+                for alias, cte in ctes.items():
+                    cte_query = cte_query.with_(alias, cte.this, recursive=recursive)
+
+                with self._concurrent_render_context():
+                    # Similar to the model's query, we render the CTE query under the locked context
+                    # so that the execution (fetchdf) can continue concurrently between the threads
+                    sql = cte_query.sql(
+                        self._test_adapter_dialect, pretty=self.engine_adapter._pretty_sql
+                    )
+
+                actual = self._execute(sql)
                 expected = self._create_df(values, columns=cte_query.named_selects, partial=partial)
 
                 self.assert_equal(expected, actual, sort=sort, partial=partial)
 
     def runTest(self) -> None:
-        query = self._render_model_query()
+        with self._concurrent_render_context():
+            # Render the model's query and generate the SQL under the locked context so that
+            # execution (fetchdf) can continue concurrently between the threads
+            query = self._render_model_query()
+            sql = query.sql(self._test_adapter_dialect, pretty=self.engine_adapter._pretty_sql)
 
-        self.test_ctes(
-            {
-                self._normalize_model_name(cte.alias, with_default_catalog=False): cte
-                for cte in query.ctes
-            }
-        )
+        with_clause = query.args.get("with")
+
+        if with_clause:
+            self.test_ctes(
+                {
+                    self._normalize_model_name(cte.alias, with_default_catalog=False): cte
+                    for cte in query.ctes
+                },
+                recursive=with_clause.recursive,
+            )
 
         values = self.body["outputs"].get("query")
         if values is not None:
             partial = values.get("partial")
             sort = query.args.get("order") is None
 
-            actual = self._execute(query)
+            actual = self._execute(sql)
             expected = self._create_df(values, columns=self.model.columns_to_types, partial=partial)
 
             self.assert_equal(expected, actual, sort=sort, partial=partial)
 
     def _render_model_query(self) -> exp.Query:
-        return self.model.render_query_or_raise(
-            **self.body.get("vars", {}),
+        variables = self.body.get("vars", {}).copy()
+        time_kwargs = {key: variables.pop(key) for key in TIME_KWARG_KEYS if key in variables}
+
+        query = self.model.render_query_or_raise(
+            **time_kwargs,
+            variables=variables,
             engine_adapter=self.engine_adapter,
             table_mapping={
                 name: self._test_fixture_table(name).sql() for name in self.body.get("inputs", {})
             },
             runtime_stage=RuntimeStage.TESTING,
         )
+        return query
 
 
 class PythonModelTest(ModelTest):
@@ -567,6 +753,8 @@ class PythonModelTest(ModelTest):
         path: Path | None = None,
         preserve_fixtures: bool = False,
         default_catalog: str | None = None,
+        concurrency: bool = False,
+        verbosity: Verbosity = Verbosity.DEFAULT,
     ) -> None:
         """PythonModelTest encapsulates a unit test for a Python model.
 
@@ -592,6 +780,8 @@ class PythonModelTest(ModelTest):
             path,
             preserve_fixtures,
             default_catalog,
+            concurrency,
+            verbosity,
         )
 
         self.context = TestExecutionContext(
@@ -615,12 +805,15 @@ class PythonModelTest(ModelTest):
 
     def _execute_model(self) -> pd.DataFrame:
         """Executes the python model and returns a DataFrame."""
-        time_ctx = freeze_time(self._execution_time) if self._execution_time else nullcontext()
-        with patch.dict(self._test_adapter_dialect.generator_class.TRANSFORMS, self._transforms):
-            with t.cast(AbstractContextManager, time_ctx):
-                df = next(self.model.render(context=self.context, **self.body.get("vars", {})))
-                assert not isinstance(df, exp.Expression)
-                return df if isinstance(df, pd.DataFrame) else df.toPandas()
+        import pandas as pd
+
+        with self._concurrent_render_context():
+            variables = self.body.get("vars", {}).copy()
+            time_kwargs = {key: variables.pop(key) for key in TIME_KWARG_KEYS if key in variables}
+            df = next(self.model.render(context=self.context, variables=variables, **time_kwargs))
+
+        assert not isinstance(df, exp.Expression)
+        return df if isinstance(df, pd.DataFrame) else df.toPandas()
 
 
 def generate_test(
@@ -655,6 +848,8 @@ def generate_test(
         name: The name of the test. This is inferred from the model name by default.
         include_ctes: When true, CTE fixtures will also be generated.
     """
+    import numpy as np
+
     test_name = name or f"test_{model.view_name}"
     path = path or f"{test_name}.yaml"
 
@@ -671,7 +866,7 @@ def generate_test(
     # ruamel.yaml does not support pandas Timestamps, so we must convert them to python
     # datetime or datetime.date objects based on column type
     inputs = {
-        models[dep].name: pandas_timestamp_to_pydatetime(
+        dep: pandas_timestamp_to_pydatetime(
             engine_adapter.fetchdf(query).apply(lambda col: col.map(_normalize_df_value)),
             models[dep].columns_to_types,
         )
@@ -681,7 +876,7 @@ def generate_test(
     }
     outputs: t.Dict[str, t.Any] = {"query": {}}
     variables = variables or {}
-    test_body = {"model": model.name, "inputs": inputs, "outputs": outputs}
+    test_body = {"model": model.fqn, "inputs": inputs, "outputs": outputs}
 
     if variables:
         test_body["vars"] = variables
@@ -695,20 +890,31 @@ def generate_test(
         path=fixture_path,
         default_catalog=model.default_catalog,
     )
+    if not test:
+        return
 
     test.setUp()
 
     if isinstance(model, SqlModel):
         assert isinstance(test, SqlModelTest)
         model_query = test._render_model_query()
+        with_clause = model_query.args.get("with")
 
-        if include_ctes:
+        if with_clause and include_ctes:
             ctes = {}
+            recursive = with_clause.recursive
             previous_ctes: t.List[exp.CTE] = []
+
             for cte in model_query.ctes:
                 cte_query = cte.this
-                for prev in previous_ctes:
-                    cte_query = cte_query.with_(prev.alias, prev.this)
+                cte_identifier = cte.args["alias"].this
+
+                cte_query = exp.select(*_projection_identifiers(cte_query)).from_(cte_identifier)
+
+                for prev in chain(previous_ctes, [cte]):
+                    cte_query = cte_query.with_(
+                        prev.args["alias"].this, prev.this, recursive=recursive
+                    )
 
                 cte_output = test._execute(cte_query)
                 ctes[cte.alias] = (
@@ -744,6 +950,19 @@ def generate_test(
         yaml.dump({test_name: test_body}, file)
 
 
+def _projection_identifiers(query: exp.Query) -> t.List[str | exp.Identifier]:
+    identifiers: t.List[str | exp.Identifier] = []
+    for select in query.selects:
+        if isinstance(select, exp.Alias):
+            identifiers.append(select.args["alias"])
+        elif isinstance(select, exp.Column):
+            identifiers.append(select.this)
+        else:
+            identifiers.append(select.output_name)
+
+    return identifiers
+
+
 def _raise_if_unexpected_columns(
     expected_cols: t.Collection[str], actual_cols: t.Collection[str]
 ) -> None:
@@ -758,6 +977,9 @@ def _raise_if_unexpected_columns(
 
 def _row_difference(left: pd.DataFrame, right: pd.DataFrame) -> pd.DataFrame:
     """Returns all rows in `left` that don't appear in `right`."""
+    import numpy as np
+    import pandas as pd
+
     rows_missing_from_right = []
 
     # `None` replaces `np.nan` because `np.nan != np.nan` and this would affect the mapping lookup
@@ -776,12 +998,14 @@ def _row_difference(left: pd.DataFrame, right: pd.DataFrame) -> pd.DataFrame:
 
 def _raise_error(msg: str, path: Path | None = None) -> None:
     if path:
-        raise TestError(f"{msg} at {path}")
-    raise TestError(msg)
+        raise TestError(f"Failed to run test at {path}:\n{msg}")
+    raise TestError(f"Failed to run test:\n{msg}")
 
 
 def _normalize_df_value(value: t.Any) -> t.Any:
     """Normalize data in a pandas dataframe so ruamel and sqlglot can deal with it."""
+    import numpy as np
+
     if isinstance(value, (list, np.ndarray)):
         return [_normalize_df_value(v) for v in value]
     if isinstance(value, dict):
@@ -791,3 +1015,43 @@ def _normalize_df_value(value: t.Any) -> t.Any:
             return {k: _normalize_df_value(v) for k, v in zip(value["key"], value["value"])}
         return {k: _normalize_df_value(v) for k, v in value.items()}
     return value
+
+
+def _split_df_by_column_pairs(df: pd.DataFrame, pairs_per_chunk: int = 4) -> t.List[pd.DataFrame]:
+    """Split a dataframe into chunks of column pairs.
+
+    Args:
+        df: The dataframe to split
+        pairs_per_chunk: Number of column pairs per chunk (default: 4)
+
+    Returns:
+        List of dataframes, each containing an even number of columns
+    """
+    total_columns = len(df.columns)
+
+    # If we have fewer columns than pairs_per_chunk * 2, return the original df
+    if total_columns <= pairs_per_chunk * 2:
+        return [df]
+
+    # Calculate number of chunks needed to split columns evenly
+    num_chunks = (total_columns + (pairs_per_chunk * 2 - 1)) // (pairs_per_chunk * 2)
+
+    # Calculate columns per chunk to ensure equal distribution
+    # We round down to nearest even number to ensure each chunk has even columns
+    columns_per_chunk = (total_columns // num_chunks) & ~1  # Round down to nearest even number
+    remainder = total_columns - (columns_per_chunk * num_chunks)
+
+    chunks = []
+    start_idx = 0
+
+    # Distribute columns evenly across chunks
+    for i in range(num_chunks):
+        # Add 2 columns to early chunks if there's a remainder
+        # This ensures we always add pairs of columns
+        extra = 2 if i < remainder // 2 else 0
+        end_idx = start_idx + columns_per_chunk + extra
+        chunk = df.iloc[:, start_idx:end_idx]
+        chunks.append(chunk)
+        start_idx = end_idx
+
+    return chunks
