@@ -6,6 +6,7 @@ import pytest
 
 from sqlmesh.core.config import ModelDefaultsConfig
 from sqlmesh.dbt.basemodel import Dependencies
+from sqlmesh.dbt.common import ModelAttrs
 from sqlmesh.dbt.context import DbtContext
 from sqlmesh.dbt.manifest import ManifestHelper, _convert_jinja_test_to_macro
 from sqlmesh.dbt.profile import Profile
@@ -33,8 +34,12 @@ def test_manifest_helper(caplog):
     assert models["top_waiters"].dependencies == Dependencies(
         refs={"sushi.waiter_revenue_by_day", "waiter_revenue_by_day"},
         variables={"top_waiters:revenue", "top_waiters:limit"},
-        model_attrs={"columns", "config"},
-        macros=[MacroReference(name="ref"), MacroReference(name="var")],
+        model_attrs=ModelAttrs(attrs={"columns", "config"}),
+        macros=[
+            MacroReference(name="get_top_waiters_limit"),
+            MacroReference(name="ref"),
+            MacroReference(name="var"),
+        ],
     )
     assert models["top_waiters"].materialized == "view"
     assert models["top_waiters"].dialect_ == "postgres"
@@ -76,7 +81,7 @@ def test_manifest_helper(caplog):
         macros=[MacroReference(name="ref")],
     )
     assert waiter_as_customer_by_day_config.materialized == "incremental"
-    assert waiter_as_customer_by_day_config.incremental_strategy == "delete+insert"
+    assert waiter_as_customer_by_day_config.incremental_strategy == "incremental_by_time_range"
     assert waiter_as_customer_by_day_config.cluster_by == ["ds"]
     assert waiter_as_customer_by_day_config.time_column == "ds"
 
@@ -98,7 +103,7 @@ def test_manifest_helper(caplog):
         has_dynamic_var_names=True,
     )
     assert waiter_revenue_by_day_config.materialized == "incremental"
-    assert waiter_revenue_by_day_config.incremental_strategy == "delete+insert"
+    assert waiter_revenue_by_day_config.incremental_strategy == "incremental_by_time_range"
     assert waiter_revenue_by_day_config.cluster_by == ["ds"]
     assert waiter_revenue_by_day_config.time_column == "ds"
     assert waiter_revenue_by_day_config.dialect_ == "bigquery"
@@ -181,7 +186,7 @@ def test_variable_override():
         profile.target,
         model_defaults=ModelDefaultsConfig(start="2020-01-01"),
     )
-    assert helper.models()["top_waiters"].limit_value == 10
+    assert helper.models()["top_waiters"].limit_value.strip() == "10"
 
     helper = ManifestHelper(
         project_path,
